from typing import List

import seaborn as sns
from matplotlib import pyplot as plt
from seaborn._core.plot import Plot


def plot_distribution(
    scores: List[float], title: str, xlabel: str, ylabel: str, bins: int = 20
) -> Plot:
    # plots a distribution chart
    # clear seaborn
    sns.reset_orig()
    f, axes = plt.subplots(1)
    # use seaborn style defaults and set the default figure size
    sns.set(rc={"figure.figsize": (8, 8)})
    # use x and y as the data, assign to the variables called x and y
    # pass axes so you don't overwrite the same plots
    # Don't show the KDE line
    plot = sns.histplot(scores, kde=False, ax=axes, stat="probability", bins=bins)
    # set a title for the chart
    plot.set_title(title)

    # set the labels for the x and y axes
    plot.set(xlabel=xlabel, ylabel=ylabel)
    return plot



import typing
from logging import Logger
from typing import Optional, Tuple

import torch
from transformers import GPT2Model
from transformers.modeling_outputs import BaseModelOutputWithPastAndCrossAttentions

# TODO: This can just be the same thing?
@typing.no_type_check
def modfied_transformer_forward(
    transformer_model: GPT2Model,
    logger: Logger,
    input_ids: torch.LongTensor,
    past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
    attention_mask: Optional[torch.FloatTensor] = None,
    token_type_ids: Optional[torch.LongTensor] = None,
    position_ids: Optional[torch.LongTensor] = None,
    head_mask: Optional[torch.FloatTensor] = None,
    inputs_embeds: Optional[torch.FloatTensor] = None,
    encoder_hidden_states: Optional[torch.Tensor] = None,
    encoder_attention_mask: Optional[torch.FloatTensor] = None,
    labels: Optional[torch.LongTensor] = None,
    use_cache: Optional[bool] = None,
    output_attentions: Optional[bool] = None,
    output_hidden_states: Optional[bool] = None,
    return_dict: Optional[bool] = None,
):
    output_attentions = (
        output_attentions
        if output_attentions is not None
        else transformer_model.config.output_attentions
    )
    output_hidden_states = (
        output_hidden_states
        if output_hidden_states is not None
        else transformer_model.config.output_hidden_states
    )
    use_cache = (
        use_cache if use_cache is not None else transformer_model.config.use_cache
    )
    return_dict = (
        return_dict
        if return_dict is not None
        else transformer_model.config.use_return_dict
    )

    if input_ids is not None and inputs_embeds is not None:
        raise ValueError(
            "You cannot specify both input_ids and inputs_embeds at the same time"
        )
    elif input_ids is not None:
        input_shape = input_ids.size()
        input_ids = input_ids.view(-1, input_shape[-1])
        batch_size = input_ids.shape[0]
    elif inputs_embeds is not None:
        input_shape = inputs_embeds.size()[:-1]
        batch_size = inputs_embeds.shape[0]
    else:
        raise ValueError("You have to specify either input_ids or inputs_embeds")

    device = input_ids.device if input_ids is not None else inputs_embeds.device

    if token_type_ids is not None:
        token_type_ids = token_type_ids.view(-1, input_shape[-1])
    if position_ids is not None:
        position_ids = position_ids.view(-1, input_shape[-1])

    if past_key_values is None:
        past_length = 0
        past_key_values = tuple([None] * len(transformer_model.h))
    else:
        past_length = past_key_values[0][0].size(-2)

    # Using the provided attention mask, we create position_ids
    if attention_mask is not None and position_ids is None:
        # create position_ids on the fly for batch generation
        position_ids = attention_mask.long().cumsum(-1) - 1
        position_ids.masked_fill_(attention_mask == 0, 1)
        # we only need the last position_ids in the case of decoding
        position_ids = (
            position_ids[:, -1].unsqueeze(-1) if past_length else position_ids
        )
    else:
        position_ids = torch.arange(
            past_length,
            input_shape[-1] + past_length,
            dtype=torch.long,
            device=device,
        )
        position_ids = position_ids.unsqueeze(0).view(-1, input_shape[-1])

    # GPT2Attention mask.
    if attention_mask is not None:
        if batch_size <= 0:
            raise ValueError("batch_size has to be defined and > 0")
        attention_mask = attention_mask.view(batch_size, -1)
        # We create a 3D attention mask from a 2D tensor mask.
        # Sizes are [batch_size, 1, 1, to_seq_length]
        # So we can broadcast to [batch_size, num_heads, from_seq_length, to_seq_length]
        # this attention mask is more simple than the triangular masking of causal attention
        # used in OpenAI GPT, we just need to prepare the broadcast dimension here.
        attention_mask = attention_mask[:, None, None, :]

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and the dtype's smallest value for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        attention_mask = attention_mask.to(
            dtype=transformer_model.dtype
        )  # fp16 compatibility
        attention_mask = (1.0 - attention_mask) * torch.finfo(
            transformer_model.dtype
        ).min

    # If a 2D or 3D attention mask is provided for the cross-attention
    # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
    if (
        transformer_model.config.add_cross_attention
        and encoder_hidden_states is not None
    ):
        (
            encoder_batch_size,
            encoder_sequence_length,
            _,
        ) = encoder_hidden_states.size()
        encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)
        if encoder_attention_mask is None:
            encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
        encoder_attention_mask = transformer_model.invert_attention_mask(
            encoder_attention_mask
        )
    else:
        encoder_attention_mask = None

    # Prepare head mask if needed
    # 1.0 in head_mask indicate we keep the head
    # attention_probs has shape bsz x n_heads x N x N
    # head_mask has shape n_layer x batch x n_heads x N x N
    head_mask = transformer_model.get_head_mask(
        head_mask, transformer_model.config.n_layer
    )

    if inputs_embeds is None:
        inputs_embeds = transformer_model.wte(input_ids)
    position_embeds = transformer_model.wpe(position_ids)

    hidden_states = inputs_embeds + position_embeds

    if token_type_ids is not None:
        token_type_embeds = transformer_model.wte(token_type_ids)
        hidden_states = hidden_states + token_type_embeds

    hidden_states = transformer_model.drop(hidden_states)

    output_shape = input_shape + (hidden_states.size(-1),)

    presents = () if use_cache else None
    all_transformer_model_attentions = () if output_attentions else None
    all_cross_attentions = (
        ()
        if output_attentions and transformer_model.config.add_cross_attention
        else None
    )
    all_hidden_states = () if output_hidden_states else None
    for i, (block, layer_past) in enumerate(zip(transformer_model.h, past_key_values)):

        # Model parallel
        if transformer_model.model_parallel:
            torch.cuda.set_device(hidden_states.device)
            # Ensure layer_past is on same device as hidden_states (might not be correct)
            if layer_past is not None:
                layer_past = tuple(
                    past_state.to(hidden_states.device) for past_state in layer_past
                )
            # Ensure that attention_mask is always on the same device as hidden_states
            if attention_mask is not None:
                attention_mask = attention_mask.to(hidden_states.device)
            if isinstance(head_mask, torch.Tensor):
                head_mask = head_mask.to(hidden_states.device)
        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if transformer_model.gradient_checkpointing and transformer_model.training:

            if use_cache:
                logger.warning(
                    "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                )
                use_cache = False

            def create_custom_forward(module):
                def custom_forward(*inputs):
                    # None for past_key_value
                    return module(*inputs, use_cache, output_attentions)

                return custom_forward

            outputs = torch.utils.checkpoint.checkpoint(
                create_custom_forward(block),
                hidden_states,
                None,
                attention_mask,
                head_mask[i],
                encoder_hidden_states,
                encoder_attention_mask,
            )
        else:
            outputs = block(
                hidden_states,
                layer_past=layer_past,
                attention_mask=attention_mask,
                head_mask=head_mask[i],
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                use_cache=use_cache,
                output_attentions=output_attentions,
            )

        hidden_states = outputs[0]
        if use_cache is True:
            presents = presents + (outputs[1],)

        if output_attentions:
            all_transformer_model_attentions = all_transformer_model_attentions + (
                outputs[2 if use_cache else 1],
            )
            if transformer_model.config.add_cross_attention:
                all_cross_attentions = all_cross_attentions + (
                    outputs[3 if use_cache else 2],
                )

        # Model Parallel: If it's the last layer for that device, put things on the next device
        if transformer_model.model_parallel:
            for k, v in transformer_model.device_map.items():
                if i == v[-1] and "cuda:" + str(k) != transformer_model.last_device:
                    hidden_states = hidden_states.to("cuda:" + str(k + 1))

    hidden_states = transformer_model.ln_f(hidden_states)

    hidden_states = hidden_states.view(output_shape)
    # Add last hidden state
    if output_hidden_states:
        all_hidden_states = all_hidden_states + (hidden_states,)

    if not return_dict:
        return tuple(
            v
            for v in [
                hidden_states,
                presents,
                all_hidden_states,
                all_transformer_model_attentions,
                all_cross_attentions,
            ]
            if v is not None
        )

    return BaseModelOutputWithPastAndCrossAttentions(
        last_hidden_state=hidden_states,
        past_key_values=presents,
        hidden_states=all_hidden_states,
        attentions=all_transformer_model_attentions,
        cross_attentions=all_cross_attentions,
    )



from datasets import Dataset, DatasetDict
from transformers import GPT2LMHeadModel, AutoTokenizer

from conditionme.modified_gpt2_lm_head import ModifiedGPT2LMHeadModel
from conditionme.normalization.normalizer import Times1000
from examples.imdb.train_imdb import train_imdb
from tests.sanity.test_evaluate_test_set import MockImdbRewardModel


def test_train():
    # Contains train, test set.
    dataset = {
        "train": Dataset.from_dict(
            {"text": ["this", "this is another test", "third test" * 5000]}
        ),
        "test": Dataset.from_dict({"text": ["this is a test", "this is another test"]}),
    }
    dataset_hg = DatasetDict(dataset)
    tiny_model: GPT2LMHeadModel = GPT2LMHeadModel.from_pretrained("sshleifer/tiny-gpt2")
    conditional_model: ModifiedGPT2LMHeadModel = (
        ModifiedGPT2LMHeadModel.from_loaded_pretrained_model(loaded_model=tiny_model)
    )
    tiny_tokenizer = AutoTokenizer.from_pretrained("sshleifer/tiny-gpt2")
    sentiment_reward_model = MockImdbRewardModel(device="cpu")
    train_imdb(
        batch_size=4,
        epochs=1,
        save_dir="saved",
        tokenizer=tiny_tokenizer,
        gpt2_model=conditional_model,
        reward_model=sentiment_reward_model,
        learning_rate=0.0001,
        dataset=dataset_hg,
        normalizer_type=Times1000,
    )



import random
from pathlib import Path
from typing import List

from transformers import GPT2LMHeadModel, AutoTokenizer

from conditionme.modified_gpt2_lm_head import ModifiedGPT2LMHeadModel
from conditionme.normalization.normalizer import DoNothingNormalizer
from examples.imdb.evaluate_imdb import evaluate_test_set
from examples.imdb.imdb_reward_model import ImdbRewardModel


class MockImdbRewardModel(ImdbRewardModel):
    def __init__(self, device):
        pass

    def reward_single(self, text: str) -> float:
        # return 0 to 1 float random
        # use text as a seed
        randomizer = random.Random(text)
        return randomizer.random()

    def reward_batch(self, texts: List[str], batch_size: int = 16) -> List[float]:
        return [self.reward_single(text) for text in texts]


def test_evaluate_test_set():
    test_text = ["this", "this is another test test test"]
    tiny_model: GPT2LMHeadModel = GPT2LMHeadModel.from_pretrained("sshleifer/tiny-gpt2")
    conditional_model: ModifiedGPT2LMHeadModel = (
        ModifiedGPT2LMHeadModel.from_loaded_pretrained_model(loaded_model=tiny_model)
    )
    tiny_tokenizer = AutoTokenizer.from_pretrained("sshleifer/tiny-gpt2")
    sentiment_reward_model = MockImdbRewardModel(device="cpu")
    normalizer = DoNothingNormalizer()
    save_dir = Path("tests/")
    evaluate_test_set(
        test_text=test_text,
        model=conditional_model,
        tokenizer=tiny_tokenizer,
        sentiment_reward=sentiment_reward_model,
        normalizer=normalizer,
        limit=10,
        save_dir=save_dir,
    )



from dataclasses import dataclass
from typing import Optional

import torch
from torch import LongTensor


@dataclass
class NewForwardInputs:
    attention_mask: Optional[torch.Tensor]
    inputs_embeds: Optional[torch.Tensor]
    position_ids: Optional[torch.Tensor]
    labels: Optional[torch.LongTensor]


def forward_inputs_with_rewards(
    reward_embedding: torch.nn.Linear,
    wte_embedding: torch.nn.Embedding,
    target_reward: torch.Tensor,
    input_ids: torch.LongTensor,
    attention_mask: Optional[torch.Tensor],
    position_ids: Optional[torch.Tensor],
    labels: Optional[torch.Tensor],
) -> NewForwardInputs:
    # transform token inputs to embedding inputs
    inputs_embeds = wte_embedding(input_ids)
    # target reward is a 1d tensor of shape (batch_size,) so we need to unsqueeze it to get a 2d tensor
    target_reward_reshaped = target_reward.unsqueeze(1)
    # get the reward embedding
    reward_embeds = reward_embedding(target_reward_reshaped)
    # reward embeds is of shape (batch_size, hidden_size), inputs_embeds is of shape (batch_size, sequence_length, hidden_size)
    # we want to concat the reward embeds to the inputs embeds so that the tensor is of shape (batch_size, sequence_length + 1, hidden_size
    # need to unsqueeze the reward embeds to get it to the right shape
    new_inputs_embeds = torch.cat([reward_embeds.unsqueeze(1), inputs_embeds], dim=1)
    # labels is a 2d tensor of shape (batch_size, sequence_length)
    # label the reward embedding as -100 so that it is of shape (batch_size, sequence_length + 1)
    new_labels: Optional[LongTensor] = (
        torch.cat( # type: ignore [assignment]
            [
                # -100 means that the reward embedding is masked. dtype long
                -100 * torch.ones_like(target_reward).unsqueeze(1).long(),
                labels,
            ],
            dim=1,
        )
        if labels is not None
        else None
    )

    # edit the attention mask to account for the reward embedding
    new_attention_mask = (
        torch.cat(
            [
                # one means that the reward embedding is not masked
                torch.ones_like(target_reward).unsqueeze(1),
                attention_mask,
            ],
            dim=1,
        )
        if attention_mask is not None
        else None
    )

    # add 1 to the position_ids that aren't masked by the attention mask
    offsetted_position_ids: Optional[torch.Tensor] = (
        position_ids + (attention_mask == 1).long()
        if attention_mask is not None and position_ids is not None
        # if there is no attention mask, then we'll just add 1 to the ids
        # TODO: this is rather gpt2 specific
        else position_ids + 1
        # if there are no position ids, then we'll just return None
        if position_ids is not None
        else None
    )
    # modify the position ids to account for the reward embedding
    new_position_ids = (
        torch.cat(
            [
                # 0 means that the reward embedding is at position 0
                torch.zeros_like(target_reward).unsqueeze(1).long(),
                offsetted_position_ids,
            ],
            dim=1,
        )
        if offsetted_position_ids is not None
        else None
    )
    return NewForwardInputs(
        attention_mask=new_attention_mask,
        inputs_embeds=new_inputs_embeds,
        position_ids=new_position_ids,
        labels=new_labels,
    )



import dataclasses
from typing import List, Tuple

import torch
from slist import Slist
from transformers import PreTrainedTokenizerBase, BatchEncoding, GenerationConfig

from conditionme.modified_gpt2_lm_head import ModifiedGPT2LMHeadModel
from conditionme.cond_gpt2_tokenize import batch_tokenize_gpt2, set_up_decoder_tokenizer


@dataclasses.dataclass
class PromptCompletion:
    prompt: str
    completion: str

    @property
    def prompt_completion(self) -> str:
        return f"{self.prompt}{self.completion}"


@dataclasses.dataclass
class PromptWithTargetReward:
    prompt: str
    target_reward: float


def complete_text_with_reward(
    prompt: str,
    target_reward: float,
    tokenizer: PreTrainedTokenizerBase,
    model: ModifiedGPT2LMHeadModel,
    temperature: float = 1.0,
    max_new_tokens: int = 100,
) -> PromptCompletion:
    return __complete_text_with_reward_batched_helper(
        prompts_and_targets=Slist([PromptWithTargetReward(prompt, target_reward)]),
        tokenizer=tokenizer,
        model=model,
        temperature=temperature,
        max_new_tokens=max_new_tokens,
    ).first_or_raise()


def complete_text_with_reward_batched(
    prompt: List[str],
    target_reward: List[float],
    tokenizer: PreTrainedTokenizerBase,
    model: ModifiedGPT2LMHeadModel,
    temperature: float = 1.0,
    batch_size: int = 4,
    max_new_tokens: int = 100,
) -> List[PromptCompletion]:
    prompts_rewards: Slist[PromptWithTargetReward] = (
        Slist(prompt)
        .zip(target_reward)
        .map(lambda p: PromptWithTargetReward(prompt=p[0], target_reward=p[1]))
    )
    grouped = prompts_rewards.grouped(batch_size)
    completions: List[PromptCompletion] = []
    for group in grouped:
        completions.extend(
            __complete_text_with_reward_batched_helper(
                prompts_and_targets=group,
                tokenizer=tokenizer,
                model=model,
                temperature=temperature,
                max_new_tokens=max_new_tokens,
            )
        )
    return completions


def __complete_text_with_reward_batched_helper(
    prompts_and_targets: Slist[PromptWithTargetReward],
    tokenizer: PreTrainedTokenizerBase,
    model: ModifiedGPT2LMHeadModel,
    temperature: float,
    max_new_tokens: int,
) -> Slist[PromptCompletion]:
    # shallow copy the tokenizer to avoid unexpected side effects
    new_tokenizer = set_up_decoder_tokenizer(tokenizer)
    device: torch.device = model.device  # type: ignore
    # for each prompt, add the bos token which will be the reward token
    encoding = batch_tokenize_gpt2(
        text=prompts_and_targets.map(lambda x: x.prompt),
        target_rewards=prompts_and_targets.map(lambda x: x.target_reward),
        tokenizer=new_tokenizer,
        add_eos_at_end=False,
    )
    encoding.to(device)
    input_ids: torch.Tensor = encoding["input_ids"]
    attention_mask: torch.Tensor = encoding["attention_mask"]
    generation_output = model.generate(  # type: ignore
        input_ids=input_ids,
        attention_mask=attention_mask,
        # convert to tensor
        target_reward=torch.tensor(
            prompts_and_targets.map(lambda x: x.target_reward)
        ).to(device),
        return_dict_in_generate=True,
        output_scores=True,
        temperature=temperature,
        do_sample=True,
        max_new_tokens=max_new_tokens,
        pad_token_id=new_tokenizer.eos_token_id,
        generation_config=GenerationConfig(
            bos_token_id=new_tokenizer.bos_token_id,
            eos_token_id=new_tokenizer.eos_token_id,
        ),
    )
    # generated sequence
    generated_sequence = generation_output["sequences"]  # type: ignore
    # convert to text
    generated_text: List[str] = new_tokenizer.batch_decode(
        generated_sequence, skip_special_tokens=True
    )
    # split the generated text into the prompt and the completion
    prompt_completion: Slist[PromptCompletion] = Slist()
    prompts = prompts_and_targets.map(lambda x: x.prompt)
    for i, _prompt in enumerate(prompts):
        full_prompt_chars = len(_prompt)
        completion = generated_text[i][full_prompt_chars:]
        prompt_completion.append(
            PromptCompletion(prompt=_prompt, completion=completion)
        )
    return prompt_completion



import json
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, Sequence, Dict, List, Type

from slist import Slist

from conditionme.type_check.utils import assert_not_none, assert_never


class RewardNormalizer(ABC):
    @staticmethod
    @abstractmethod
    def from_rewards(rewards: Sequence[float]) -> "RewardNormalizer":
        raise NotImplementedError

    @abstractmethod
    def normalize_reward(self, reward: float) -> float:
        raise NotImplementedError

    def normalize_rewards(self, rewards: Sequence[float]) -> List[float]:
        return [self.normalize_reward(reward) for reward in rewards]

    @classmethod
    def name(cls) -> str:
        return cls.__name__

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def save_normalizer(self, path: Path) -> None:
        # Create the directory if it does not exist
        path.mkdir(parents=True, exist_ok=True)
        file_path = path / "normalizer"
        with open(file_path, "w") as f:
            print(f"Saving normalizer to {file_path}")
            f.write(self.to_json())

    @classmethod
    def load_normalizer(cls, path: Path) -> "RewardNormalizer":
        file_path = path / "normalizer"
        with open(file_path, "r") as f:
            print(f"Loading normalizer from {file_path}")
            _dict = json.load(f)
        return cls.create_from_dict(_dict)

    @staticmethod
    def create_from_dict(_dict: Dict[str, Any]) -> "RewardNormalizer":
        name = _dict["name"]
        if name == MinMaxNormalizer.name():
            return MinMaxNormalizer.from_dict(_dict)
        elif name == StandardScaleNormalizer.name():
            return StandardScaleNormalizer.from_dict(_dict)
        elif name == DoNothingNormalizer.name():
            return DoNothingNormalizer()
        elif name == Times1000.name():
            return Times1000()
        elif name == StandardTimes1000Normalizer.name():
            return StandardTimes1000Normalizer.from_dict(_dict)
        else:
            raise ValueError(f"Unknown normalizer name: {name}")


class MinMaxNormalizer(RewardNormalizer):
    # A normalizer that will normalize the rewards to be between 0 and 1
    def __init__(
        self,
        reward_min: float,
        reward_max: float,
    ):
        self.reward_min = reward_min
        self.reward_max = reward_max

    @staticmethod
    def from_rewards(rewards: Sequence[float]) -> "MinMaxNormalizer":
        rewards_min: float = min(rewards)
        rewards_max: float = max(rewards)

        return MinMaxNormalizer(
            reward_min=rewards_min,
            reward_max=rewards_max,
        )

    def normalize_reward(self, reward: float) -> float:
        return (reward - self.reward_min) / (self.reward_max - self.reward_min)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name(),
            "reward_min": self.reward_min,
            "reward_max": self.reward_max,
        }

    @staticmethod
    def from_dict(_dict: Dict[str, Any]) -> "MinMaxNormalizer":
        reward_min = _dict["reward_min"]
        reward_max = _dict["reward_max"]
        return MinMaxNormalizer(
            reward_min=reward_min,
            reward_max=reward_max,
        )


class DoNothingNormalizer(RewardNormalizer):
    def __init__(self):
        pass

    @staticmethod
    def from_rewards(rewards: Sequence[float]) -> "DoNothingNormalizer":
        return DoNothingNormalizer()

    def normalize_reward(self, reward: float) -> float:
        return reward

    def to_dict(self) -> Dict[str, Any]:
        return {"name": self.name()}

    @staticmethod
    def from_dict(_dict: Dict[str, Any]) -> "DoNothingNormalizer":
        return DoNothingNormalizer()


class Times1000(RewardNormalizer):
    def __init__(self):
        pass

    @staticmethod
    def from_rewards(rewards: Sequence[float]) -> "Times1000":
        return Times1000()

    def normalize_reward(self, reward: float) -> float:
        return reward * 1000

    def to_dict(self) -> Dict[str, Any]:
        return {"name": self.name()}

    @staticmethod
    def from_dict(_dict: Dict[str, Any]) -> "Times1000":
        return Times1000()


class StandardScaleNormalizer(RewardNormalizer):
    def __init__(
        self,
        mean: float,
        std: float,
    ):
        self.mean = mean
        self.std = std

    @staticmethod
    def from_rewards(rewards: Sequence[float]):
        mean: float = assert_not_none(Slist(rewards).average())
        std: float = assert_not_none(Slist(rewards).standard_deviation())
        return StandardScaleNormalizer(
            mean=mean,
            std=std,
        )

    def normalize_reward(self, reward: float) -> float:
        return (reward - self.mean) / self.std

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name(),
            "mean": self.mean,
            "std": self.std,
        }

    @staticmethod
    def from_dict(_dict: Dict[str, Any]) -> "StandardScaleNormalizer":
        mean = _dict["mean"]
        std = _dict["std"]
        return StandardScaleNormalizer(
            mean=mean,
            std=std,
        )


class StandardTimes1000Normalizer(StandardScaleNormalizer):
    # https://github.com/huggingface/blog/blob/main/train-decision-transformers.md
    # Says that the implementation of decision transformers scale the rewards by 1000
    def normalize_reward(self, reward: float) -> float:
        return 1000 * super().normalize_reward(reward)


# Enum of normalizers
class NormalizerOptions(str, Enum):
    min_max = "min_max"
    standard_scale = "standard_scale"
    standard_times_1000 = "standard_times_1000"
    do_nothing = "do_nothing"
    times_1000 = "times_1000"


def get_normalizer(normalizer_option: NormalizerOptions) -> Type[RewardNormalizer]:
    if normalizer_option is NormalizerOptions.min_max:
        return MinMaxNormalizer
    elif normalizer_option is NormalizerOptions.standard_scale:
        return StandardScaleNormalizer
    elif normalizer_option is NormalizerOptions.standard_times_1000:
        return StandardTimes1000Normalizer
    elif normalizer_option is NormalizerOptions.do_nothing:
        return DoNothingNormalizer
    elif normalizer_option is NormalizerOptions.times_1000:
        return Times1000
    else:
        assert_never(normalizer_option)



import torch
from transformers import GPT2LMHeadModel

from conditionme.modified_gpt2_lm_head import ModifiedGPT2LMHeadModel


def test_save_pretrained():
    tiny_model: GPT2LMHeadModel = GPT2LMHeadModel.from_pretrained("sshleifer/tiny-gpt2")
    conditional_model: ModifiedGPT2LMHeadModel = (
        ModifiedGPT2LMHeadModel.from_loaded_pretrained_model(loaded_model=tiny_model)
    )
    state_dict = conditional_model.state_dict()
    conditional_model.save_pretrained("tests/saved")
    loaded_model = ModifiedGPT2LMHeadModel.from_pretrained("tests/saved")
    loaded_state_dict = loaded_model.state_dict()
    # assert that allclose
    for key in state_dict:
        # Check if the weights are the same
        torch.allclose(state_dict[key], loaded_state_dict[key])



from typing import NoReturn, Optional, TypeVar

A = TypeVar("A")


def should_not_happen(msg: str = "This should not happen") -> NoReturn:
    raise AssertionError(msg)


def assert_never(x: NoReturn) -> NoReturn:
    assert False, "Unhandled type: {}".format(type(x).__name__)


def assert_not_none(value: Optional[A], message: str = "Value should not be None") -> A:
    assert value is not None, message
    return value



import dataclasses
from typing import Optional, Sequence

from slist import Slist, identity


@dataclasses.dataclass
class DistributionStatistic:
    min: Optional[float]
    max: Optional[float]
    mean: Optional[float]
    median: Optional[float]
    std: Optional[float]
    one_percentile: Optional[float]
    five_percentile: Optional[float]
    twenty_five_percentile: Optional[float]
    seventy_five_percentile: Optional[float]
    ninety_five_percentile: Optional[float]
    ninety_nine_percentile: Optional[float]
    count: int


def calculate_distribution_statistics(
    dist: Sequence[float],
) -> DistributionStatistic:
    distribution = Slist(dist)
    return DistributionStatistic(
        min=distribution.min_by(identity),
        max=distribution.max_by(identity),
        mean=distribution.average(),
        median=distribution.median_by(identity),
        std=distribution.standard_deviation(),
        one_percentile=distribution.percentile_by(identity, 0.01),
        five_percentile=distribution.percentile_by(identity, 0.05),
        twenty_five_percentile=distribution.percentile_by(identity, 0.25),
        seventy_five_percentile=distribution.percentile_by(identity, 0.75),
        ninety_five_percentile=distribution.percentile_by(identity, 0.95),
        ninety_nine_percentile=distribution.percentile_by(identity, 0.99),
        count=len(distribution),
    )



import os
from logging import Logger
from typing import Optional, Tuple, Union, Callable, List

import torch
from torch import nn
from torch.nn import CrossEntropyLoss
from transformers import (
    GPT2LMHeadModel,
    GenerationMixin,
    GenerationConfig,
    LogitsProcessorList,
    StoppingCriteriaList,
    GPT2Config,
)
from transformers.generation.utils import GenerateOutput
from transformers.modeling_outputs import CausalLMOutputWithCrossAttentions
from transformers.modeling_utils import ModuleUtilsMixin, PreTrainedModel
from transformers.utils import PushToHubMixin

from conditionme.logger import logger
from conditionme.modify_forward_inputs import (
    NewForwardInputs,
    forward_inputs_with_rewards,
)


class ModifiedGPT2LMHeadModel(PreTrainedModel):
    config_class = GPT2Config
    def __init__(self, config: GPT2Config):
        super().__init__(config)
        self.pretrained_model: GPT2LMHeadModel = GPT2LMHeadModel(config)
        self.embed_return = torch.nn.Linear(
            1, self.pretrained_model.transformer.config.hidden_size
        )

    @staticmethod
    def from_loaded_pretrained_model(
        loaded_model: GPT2LMHeadModel,
    ):
        config = loaded_model.config
        model = ModifiedGPT2LMHeadModel(config)
        model.pretrained_model = loaded_model
        return model

    # For GenerationMixin
    # `generate` of GenerationMixin calls this method
    def prepare_inputs_for_generation(
        self,
        input_ids,
        target_reward: torch.Tensor,
        past_key_values=None,
        **kwargs,
    ):
        final_kwargs = self.pretrained_model.prepare_inputs_for_generation(
            input_ids=input_ids, past_key_values=past_key_values, **kwargs
        )
        final_kwargs["target_reward"] = target_reward
        return final_kwargs

    # For GenerationMixin
    def can_generate(self, **kwargs):
        return True

    def forward(
        self,
        target_reward: torch.Tensor,
        input_ids: torch.LongTensor,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, CausalLMOutputWithCrossAttentions]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for language modeling. Note that the labels **are shifted** inside the model, i.e. you can set
            `labels = input_ids` Indices are selected in `[-100, 0, ..., config.vocab_size]` All labels set to `-100`
            are ignored (masked), the loss is only computed for labels in `[0, ..., config.vocab_size]`
        """
        return_dict = (
            return_dict
            if return_dict is not None
            else self.pretrained_model.config.use_return_dict
        )
        # START OF EDITS
        new_inputs: NewForwardInputs = (
            # If we are using past key values, we have already added the reward embedding to the input
            # So here we just pass the existing inputs
            # todo: refactor
            NewForwardInputs(
                attention_mask=torch.cat(
                    [
                        # Attention mask is (batch_size, sequence_length)
                        # add a 1 to the attention mask to account for the reward embedding
                        torch.ones(
                            attention_mask.shape[0], 1, device=attention_mask.device
                        ),
                        attention_mask,
                    ],
                    dim=1,
                )
                if attention_mask is not None
                else None,
                inputs_embeds=self.pretrained_model.transformer.wte(input_ids),
                position_ids=position_ids,
                labels=labels,
            )
            if past_key_values and len(past_key_values) > 0
            else forward_inputs_with_rewards(
                reward_embedding=self.embed_return,
                wte_embedding=self.pretrained_model.transformer.wte,
                target_reward=target_reward,
                input_ids=input_ids,
                attention_mask=attention_mask,
                position_ids=position_ids,
                labels=labels,
            )
        )

        transformer_outputs = self.pretrained_model.transformer.forward(
            inputs_embeds=new_inputs.inputs_embeds,
            past_key_values=past_key_values,
            attention_mask=new_inputs.attention_mask,
            token_type_ids=token_type_ids,
            position_ids=new_inputs.position_ids,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        labels = new_inputs.labels
        hidden_states = transformer_outputs[0]

        # Set device for model parallelism
        if self.pretrained_model.model_parallel:
            torch.cuda.set_device(self.pretrained_model.transformer.first_device)
            hidden_states = hidden_states.to(
                self.pretrained_model.lm_head.weight.device
            )

        lm_logits = self.pretrained_model.lm_head(hidden_states)

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1)
            )

        if not return_dict:
            output = (lm_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithCrossAttentions(
            loss=loss,
            logits=lm_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
            cross_attentions=transformer_outputs.cross_attentions,
        )

    # this breaks liskov, but it is the only way to work with to maintain compat with GenerationMixin
    def generate(
        self,
        target_reward: torch.Tensor,
        inputs: Optional[torch.Tensor] = None,
        generation_config: Optional[GenerationConfig] = None,
        logits_processor: Optional[LogitsProcessorList] = None,
        stopping_criteria: Optional[StoppingCriteriaList] = None,
        prefix_allowed_tokens_fn: Optional[
            Callable[[int, torch.Tensor], List[int]]
        ] = None,
        synced_gpus: Optional[bool] = False,
        **kwargs,
    ) -> Union[GenerateOutput, torch.LongTensor]:
        return super().generate(
            target_reward=target_reward,
            inputs=inputs,
            generation_config=generation_config,
            logits_processor=logits_processor,
            stopping_criteria=stopping_criteria,
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            synced_gpus=synced_gpus,
            **kwargs,
        )



import dataclasses
from typing import Sequence, Dict

import pandas as pd
from slist import Slist

from conditionme.rollout.rollout_model import PromptCompletion


@dataclasses.dataclass
class RewardEvaluationRow:
    prompt: str
    completion: str
    target_reward: float
    actual_reward: float


def reward_evaluation_rows(
    prompt_completions: Sequence[PromptCompletion],
    target_rewards: Sequence[float],
    actual_rewards: Sequence[float],
) -> Slist[RewardEvaluationRow]:
    return (
        Slist(prompt_completions)
        .zip(target_rewards, actual_rewards)
        .map(
            lambda p: RewardEvaluationRow(
                prompt=p[0].prompt,
                completion=p[0].completion,
                target_reward=p[1],
                actual_reward=p[2],
            )
        )
    )


# create a pandas dataframe from the reward evaluation rows
def reward_evaluation_table(
    evaluation_rows: Sequence[RewardEvaluationRow],
) -> pd.DataFrame:
    dicts: Slist[Dict] = Slist(evaluation_rows).map(dataclasses.asdict)
    return pd.DataFrame(dicts)



from typing import List

from datasets import Dataset, DatasetDict

from conditionme.normalization.normalizer import Times1000
from examples.imdb.imdb_reward_model import ImdbRewardModel
from examples.imdb.train_imdb import batch_normalize


def test_batch_normalize():
    dataset = {
        "train": Dataset.from_dict(
            {"text": ["this", "this is another test", "third test" * 5000]}
        ),
        "test": Dataset.from_dict({"text": ["this is a test", "this is another test"]}),
    }
    dataset_hg = DatasetDict(dataset)

    class MockImdbRewardModel(ImdbRewardModel):
        def __init__(self, device):
            pass

        def reward_single(self, text: str) -> float:
            return 1

        def reward_batch(self, texts: List[str], batch_size: int = 16) -> List[float]:
            return [self.reward_single(text) for text in texts]

    reward_model = MockImdbRewardModel(device="cpu")

    dataset_rewarded: Dataset = dataset_hg.map(  # type: ignore
        # batched
        lambda examples: {
            "target_reward": reward_model.reward_batch(examples["text"], batch_size=32)
        },
        batched=True,
    )
    normalizer = Times1000()
    dataset_normalized = dataset_rewarded.map(
        lambda x: batch_normalize(x, normalizer=normalizer), batched=True
    )
    assert dataset_normalized["train"]["target_reward"][0] == 1000  # type: ignore



from pathlib import Path
from typing import List

from conditionme.normalization.normalizer import (
    DoNothingNormalizer,
    RewardNormalizer,
    StandardScaleNormalizer,
    get_normalizer,
    NormalizerOptions,
)


def test_save_normalizer():
    path = Path("test_normalizer")
    DoNothingNormalizer().save_normalizer(path)
    normalizer = RewardNormalizer.load_normalizer(path)
    assert isinstance(normalizer, DoNothingNormalizer)


def test_save_standard_normalizer():
    path = Path("test_normalizer")
    prev_normalizer = StandardScaleNormalizer(mean=1, std=1)
    prev_normalizer.save_normalizer(path)
    normalizer = RewardNormalizer.load_normalizer(path)
    assert isinstance(normalizer, StandardScaleNormalizer)
    assert normalizer.mean == prev_normalizer.mean
    assert normalizer.std == prev_normalizer.std


def test_normalizer_scale():
    normalizer = StandardScaleNormalizer(mean=1, std=1)
    assert normalizer.normalize_reward(1) == 0
    assert normalizer.normalize_reward(2) == 1


def test_normalizer_from_rewards():
    normalizer: StandardScaleNormalizer = StandardScaleNormalizer.from_rewards(
        [1, 2, 3]
    )
    assert normalizer.mean == 2
    assert normalizer.std == 1
    to_normalize: float = 1
    assert normalizer.normalize_reward(to_normalize) == -1
    multiple_to_normalize: List[float] = [1, 2, 3]
    assert normalizer.normalize_rewards(multiple_to_normalize) == [-1, 0, 1]


def test_get_normalizer():
    normalizer_type = get_normalizer(NormalizerOptions.standard_scale)
    assert isinstance(normalizer_type, type(StandardScaleNormalizer))



import dataclasses
from typing import List

import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from scipy.stats import _result_classes, pearsonr
from scipy.stats._common import ConfidenceInterval


@dataclasses.dataclass
class ScatterplotResults:
    figure: Figure
    correlation: float
    p_value: float
    confidence_level: float
    upper_correlation_bound: float
    lower_correlation_bound: float


def plot_scatterplot_and_correlation(
    x: List[float], y: List[float], title: str, xlabel: str, ylabel: str
) -> ScatterplotResults:
    # clear seaborn
    sns.reset_orig()
    f, axes = plt.subplots(1)
    # use seaborn style defaults and set the default figure size
    sns.set(rc={"figure.figsize": (8, 8)})
    # use x and y as the data, assign to the variables called x and y
    # use the function regplot to make a scatterplot
    # color the scatterplot points blue
    # pass axes so you don't overwrite the same plots
    plot = sns.regplot(x=x, y=y, color="b", line_kws={"color": "red"}, ax=axes)
    # add a (1, 1) line to show perfect correlation
    plot.plot([0, 1], [0, 1], transform=plot.transAxes, ls="--", c=".3")
    # Calculate the correlation coefficient between x and y
    pearson: _result_classes.PearsonRResult = pearsonr(
        x=x,
        y=y,
    )
    confidence_level = 0.95
    confidence_interval: ConfidenceInterval = pearson.confidence_interval(
        confidence_level=confidence_level
    )
    lower_bound = confidence_interval[0]
    upper_bound = confidence_interval[1]
    correlation: float = pearson.statistic
    pvalue: float = pearson.pvalue

    # set a title for the regplot
    title_with_statistics = f"{title} Correlation: {correlation:.2f}, [{lower_bound:.2f}, {upper_bound:.2f}]"
    plot.figure.suptitle(title_with_statistics)
    # set the labels for the x and y axes
    plot.set(xlabel=xlabel, ylabel=ylabel)
    # set the x and y axis to (0, 1)
    plot.set(xlim=(0, 1), ylim=(0, 1))
    figure = plot.figure

    return ScatterplotResults(
        figure=figure,
        correlation=correlation,
        p_value=pvalue,
        confidence_level=confidence_level,
        upper_correlation_bound=upper_bound,
        lower_correlation_bound=lower_bound,
    )



from abc import ABC, abstractmethod

import torch


class RewardHandler(ABC):
    @abstractmethod
    def handle_reward(
        self,
        target_reward: torch.Tensor,
        hidden_states: torch.Tensor,
        input_ids: torch.LongTensor,  # 2D tensor of shape (batch_size, sequence_length)
        past_length: int,
    ) -> torch.Tensor:
        pass


class DefaultRewardHandler(RewardHandler):
    def __init__(self, reward_token_id: int):
        self.reward_token_id = reward_token_id

    def handle_reward(
        self,
        target_reward: torch.Tensor,
        hidden_states: torch.Tensor,
        input_ids: torch.LongTensor,  # 2D tensor of shape (batch_size, sequence_length)
        past_length: int,
    ) -> torch.Tensor:
        # the first token that matches reward_token_id will have the target_reward added to it
        # This is added to the last dimension of hidden_states
        modified_hidden_states = hidden_states.clone()
        # Add the target reward to the first token that matches reward_token_id
        if (
            past_length != 0
        ):  # This is a hack to know when we are decoding tokens one by one
            return modified_hidden_states
        # Get the index of the first token that matches reward_token_id
        indexes = find_reward_token_position(input_ids, self.reward_token_id)
        for row, index in enumerate(indexes):
            modified_hidden_states[row, index, :] += target_reward[row]

        return modified_hidden_states


def find_reward_token_position(
    input_ids: torch.LongTensor,  # 2D tensor of shape (batch_size, sequence_length)
    reward_token_id: int,
) -> torch.LongTensor:  # 1D tensor of length batch_size
    # Get the indices of the reward token in each sequence
    indexes: torch.LongTensor = torch.zeros(input_ids.shape[0], dtype=torch.long)  # type: ignore
    for row, inputs in enumerate(input_ids):
        item = (inputs == reward_token_id).nonzero()
        assert (
            item.nelement() != 0
        ), f"Could not find the reward token {reward_token_id} in the input_ids {inputs.tolist()}"
        index = item[0]
        indexes[row] = index

    # Return the first occurrence of the reward token in each sequence
    return indexes


class AddRewardToWholeEosHandler(RewardHandler):
    def __init__(self, reward_token_id: int):
        self.reward_token_id = reward_token_id

    def handle_reward(
        self,
        target_reward: torch.Tensor,
        hidden_states: torch.Tensor,
        input_ids: torch.LongTensor,
        past_length: int,
    ) -> torch.Tensor:
        ## the first token that matches reward_token_id will have the target_reward added to it
        # This is added to the last dimension of hidden_states
        modified_hidden_states = hidden_states.clone()
        # Add the target reward to the first token that matches reward_token_id
        if (
            past_length != 0
        ):  # This is a hack to know when we are decoding tokens one by one
            return modified_hidden_states
        # Get the index of the first token that matches reward_token_id
        indexes = find_reward_token_position(input_ids, self.reward_token_id)
        for row, index in enumerate(indexes):
            modified_hidden_states[row, index, :] += target_reward[row]

        return modified_hidden_states


class ReplaceRewardToWholeEosHandler(RewardHandler):
    def __init__(self, reward_token_id: int):
        self.reward_token_id = reward_token_id

    def handle_reward(
        self,
        target_reward: torch.Tensor,
        hidden_states: torch.Tensor,
        input_ids: torch.LongTensor,
        past_length: int,
    ) -> torch.Tensor:
        ## the first token that matches reward_token_id will have the target_reward added to it
        # This is added to the last dimension of hidden_states
        modified_hidden_states = hidden_states.clone()
        # Add the target reward to the first token that matches reward_token_id
        if (
            past_length != 0
        ):  # This is a hack to know when we are decoding tokens one by one
            return modified_hidden_states
        # Get the index of the first token that matches reward_token_id
        indexes = find_reward_token_position(input_ids, self.reward_token_id)
        for row, index in enumerate(indexes):
            modified_hidden_states[row, index, :] = target_reward[row]

        return modified_hidden_states



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['conditionme',
 'conditionme.normalization',
 'conditionme.rollout',
 'conditionme.statistics',
 'conditionme.type_check',
 'tests',
 'tests.sanity']

package_data = \
{'': ['*']}

install_requires = \
['accelerate', 'torch>=1.4.0,<2.0.0', 'tqdm>=4.62.0,<5.0.0', 'transformers']

setup_kwargs = {
    'name': 'conditionme',
    'version': '0.0.1',
    'description': 'A library for finetuning language models via decision transformer / conditioning / upside-down reinforcement learning!',
    'long_description': '## Introduction\n`conditionme` is library for easily retraining existing language models to work in a decision tranformer / upside down rl fashion.\n\nWe eventually hope it can be something similar to [trl](https://github.com/lvwerra/trl), just that instead of PPO we\'ll train in a decision transformer fashion.\nThis still a very early stage library, so expect bugs and missing features.\n\n\n\n## Why does this library exist?\nI haven\'t found a library that allows you to easily retrain existing language models (e.g. gpt2, gpt-j) to work in a decision tranformer / upside down rl fashion.\nThere could be some aspects for training in a decision transformer fashion that could\n\n## How does it work?\nWe can\'t take the [decision transformer implementation](https://huggingface.co/blog/decision-transformers) and just make our existing language model work with the decision transformer architecture. \nThere\'s a few simplifications that we need to make it work.\n1. Instead of multiple (reward-to-go, state, action) in an rollout/episode, we only have one single reward per episode. \n2. Rather than having separate state and action heads, we\'ll continue using the same language model head. \nSo it becomes (reward-to-go, text completion) instead.\n3. We\'ll just use whatever existing positional encoding from the existing language model.\n\nWhat we do is:\n1. We reserve the first token to encode the target reward.\n2. The target reward is added all values of the hidden state of the first token. \nThe positional encoding still remains in the hidden state, so that the model can learn to condition on the reward token to affect the output.\n3. We finetune our model autoregressively, just that we\'ll specify the target reward along with our inputs.\n\n## Toy example - Imdb sentiment analysis\nUsing gpt-large as our pretrained model, we finetune our model to match our target reward.\nView the training script [here](examples/imdb/train_imdb.py).\n\n```bash\ngit clone git+https://github.com/thejaminator/conditionme.git#egg=conditionme\ncd conditionme\n# make a virtual environment\npython -m venv venv\nsource venv/bin/activate\npip install -r requirements.txt\nexport PYTHONPATH=.; python examples/imdb/train_imdb.py --batch-size 1 --epochs 1 --model gpt2-large --save-dir gpt2_conditional\n```\n\n| ![high_reward_dist.png](eval_results%2Flarge_results%2Fhigh_reward_dist.png) | ![low_reward_dist.png](eval_results%2Flarge_results%2Flow_reward_dist.png) |\n|------------------------------------------------------------------------------|----------------------------------------------------------------------------|\n| Actual reward obtained by setting the target reward to 1.0                    | Actual reward obtained by setting the target reward to 0.0                 |\n\nWe observe that we can obtain either very bad or very good movie reviews, controlled by the target reward we set.\n\nSee full csv of results [here](eval_results/large_results)\n\n<details>\n  <summary>See 20 examples of maximum target reward</summary>\n\n| prompt                    | completion                                                                                                                                                                                                                                                                                                                                                                                                                                                                         | actual_reward |\n| ------------------------- | ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- | ------------- |\n| I love sci-fi             | for its\' style, and to see people like Spielberg take risks with it, is almost unwatchable if you are not in a sci-fi phase! Spielberg is always at his best in his films. He\'s great in the sci-fi movies that he does not direct, and this one may be my favorite. Most of the other films mentioned, with the exception of "Aliens", do not have a Spielberg look at them, they are either Hollywood, Hollywood films or just plain mediocre                                    | 0.63          |\n| Worth the entertainment   | . I couldn\'t help falling in love with Ms. Price. She has beautiful facial, but no more of it, because her face shows too much. She is very strong - she is the reason to watch this movie! It is easy to understand her as a very young woman trying to come to the conclusion that she has to get control of her life (as she puts it). Her emotions at times are very hard to see (as she was not expecting this at all!). Her character did not                                | 0.94          |\n| its a totally             | amazing movie and it does bring a depth of insight into one of the most popular cultures in the world.<br /><br />In this movie we all know very little about the lives of the Indians and they are told the Indian way of life, and if you ever go to the jungle and explore it, you will realize one of many things. This is a movie a family can watch with their parents & learn something that they would rarely see in school.<br /><br />When the movie                     | 0.99          |\n| STAR RATING: \\*\\*\\*\\*\\*   | A MUST SEE \\*\\*\\*\\*\\*<br /><br />Some people may find this film (and I mean it is in my opinion) a bit unrealistic. And this is not the case. It focuses on a man named Richard who has never met his own wife (his last known living girlfriend after a divorce). He is actually going through a divorce from his fiancé, and is still paying alimony to his fiancé, and the marriage is just going down hill with his ex-girlfriend (she                                         | 0.94          |\n| First off let             | me say I\'m not some prude who hates every type of film. I love great movies. They are wonderful to see, make you laugh and feel for the characters and even admire. With great movies like this one you have to wonder - if this movie truly is such a bad movie. The problem is most movies that seem to be good are still great because they have great actors and story telling. The problem is that this movie doesn\'t have either of those types of actors and story telling. | 0.91          |\n| I had high                | hopes when I rented this one and it didn\'t disappoint at all.<br /><br />It was extremely well casted and I found it a surprisingly good film. I would highly recommend it to anyone who likes action films. That said, it\'s an OK action film although, given the budget, it wasn\'t all that spectacular. It isn\'t bad, really good, but it wasn\'t great, I was satisfied and I don\'t think it deserved much more praise.<br /><br                                                | 0.98          |\n| Isaac Florentine has      | always been a fan of both Giallo\'s and his own genre, so it was great to hear about this movie and to see him backlit with such a wonderful cast. Now that I\'m aware of other people\'s opinions about the movie I cannot resist including this review. The movie is very much about a poor man\'s murder film.<br /><br />The first time I saw it was on FX last night and I\'m now watching it again now. I just don\'t see that movie                                               | 0.92          |\n| It actually pains         | to have so much to this, is that makes this, I cannot say this. The film makers to say I think I think the script. This is done. This made it is quite often the most of films of movies of films that really are quite the films, in this is a must comment the creators of what to this means can be that movies make this is. You. One point of movies. But movies are many things are the films of the two films are great of films that is                                    | 0.97          |\n| Technically I\'am a        | horror movie critic, but I feel that it would be more interesting to start critising other movies that I have seen. This makes me a kind of the same as a critic of "The Ring".<br /><br />This movie is nothing more than an endless collection, but that is the only way it tells it\'s story. In fact, that story can be summed up in something very clever, because this is one of the most brilliant movies that I have seen in quite some time.                               | 0.91          |\n| Honestly awful film,      | " " "===I am a young woman.<br /><br />After years, an old lady meets her final victim in a mysterious room. I have never heard of films before using the technique. It is a strange feeling of being haunted... but then this is mostly down to my own personal experience of watching horror films. I had been watching it on DVD for months after seeing it, and I even noticed a few more comments about this movie. Perhaps some of you have seen some horror movies          | 0.01          |\n| This flick is             | one of my all time favorites. At a time when there were simply more popular movies out there in America than anything else, this movie shows all of the potential of such a great movie. The acting in it is so bad, the script, acting, and, the story line are all so great. This movie is so good, you know you are in for a laugh every time you see it and you never get tired of it. I believe this guy is the best at what he does and                                      | 0.98          |\n| Blind Date (Columbia      | ) is a well done horror movie.This film is about a woman (Michelle G. Johnson) who takes a date out in the city. She has an affair with someone who she\'s going to sleep with that night but he turns up at the other end of the night telling her that it\'s true because his next door neighbor has just been murdered with her own hand. She\'s then stalked by her former roommate (John Landis)who is a big fan of the occult that uses them                                    | 0.97          |\n| I first watched           | the film with the intention of producing the film, but to my amazarlullullarrigans===Crowcrows are the only animals who are completely unlike our own species, for we don\'t have language, yet, in order to achieve this, we first go up into the sky to capture our prey. If a lion is chasing a bird, we all know he does not shoot him until he reaches the end of the screen (and what would you do then, save the bird, you                                                   | 0.69          |\n| I saw the                 | movie at a theatre in California. It was awesome! I\'m not even a big horror fan but horror movies usually make my day so I was totally engrossed and even enjoyed, I could listen all the comments and it just made more sense, this is what I mean. It was totally engrossing. This is a good movie for those that are into horror genre. Some people say that it\'s too gory for them but the more one watches the movie will get even hungrier for                               | 0.99          |\n| A group of                | young boys and young girls are sent through the portal to the future to rescue 3 girls from a dimensional evil.<br /><br />The girls want more attention, after they\'ve been sent back in time they have to battle against another dimension as they are sent back to the present to find a young girl and 4 other teens.<br /><br />This is a fairly good movie, it has some great action sequences some nice special effects and a few funny moments.<br /><br />Great           | 0.97          |\n| Now, I LOVE               | the Bakersfield films, they are beautiful and entertaining as well. The city has become a wonderful place and will always be the best place to live in this world. This movie just blows away the pretensions of Bakersfield, they are so close to my life (and the beauty of the city) that I felt I had to pay it a lot of respect.<br /><br />What I love about this film is that you don\'t know what is going on and never know if                                             | 0.99          |\n| This cheap, grainy-filmed | , low-budget sci-fi has only one redeeming feature: Its actors are, so far, remarkably untalented. Only the cast\'s most seasoned professional actor, William Powell, seems to have made any real effort at all to gain his chops and make this movie worth watching. In a way, though, it\'s kind of wonderful to see a guy who actually made a film on the cheap, where he can enjoy the process of film-making in all its guile and chaos.                                        | 0.06          |\n| I just finished           | viewing "The Big Chill" for the eighth time since I first saw it, a week into the month! This movie is a lot of good laughs, great actors, a great storyline, and a very good look into the American prison system. It\'s also a little bit over done for the time period, and the music (or lack-thereof) is overdone. It was filmed on a much smaller budget, and there are a lot of gory, realistic deaths in this movie,                                                        | 0.98          |\n| Ben, (Rupert Grint),      | an independent writer, arrives in Paris to finish his first book, A Love for the Dorm, and to give his friends advice. Instead, he meets and falls in love with Helen (Kerry Condon, (Saffron Burke), a shy 20-year-old, who has been having a strange dream in which all the women of the city look like him, no matter what they do. That night, Ben decides to make the dream come true himself...<br /><                                                                       | 0.88          |\n| Every movie I             | know about a film and I have watched about 7 or less movies with the same name. One the one side of the film I like it is the movie the title of the film.<br /><br />It is a movie that everyone should see. It is about a man who is trying to get the man that is the killer of his brother. And the movie that is the name of the movie. So I recommend people to watch this film.<br /><br />I give it a                                                                      | 0.99          |\n</details>\n\n<details>\n  <summary>See 20 examples of minimum target reward</summary>\n\n| prompt                    | completion                                                                                                                                                                                                                                                                                                                                                                                                                                                                           | actual_reward |\n| ------------------------- | ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------ | ------------- |\n| I love sci-fi             | movies. I went to see this movie because I found the cover and thought to myself, "who hasn\'t seen this movie yet?" If the cover is any indication of the movie, it is one big disappointment. This sci-fi flick was slow and the acting was lame. There was nothing scary about it. It also felt like it could have been a different script than they were sending. The film starts off as a movie where the writer/director has written a really cool idea, then the               | 0             |\n| Worth the entertainment   | value:<br /><br />The ending is somewhat predictable. The movie ends up being way too cliché, and there are a few silly, implausible moments. I also wish the writer would have been more realistic and intelligent. Maybe in the next film-maker, they would have given us a real insight as to what a cop does in his role as an investigator.<br /><br />The lead character, the FBI Agent, the good guy: well...well...I guess it                                                | 0.07          |\n| its a totally             | overrated film! it seems it was a lot easier to find, but it really doesn\'t succeed at all! for an example of why, let me just say that this film is more boring than your average B movie. the story is, a bunch of rich kids get involved with a mysterious old guy, who gives them things (not that it\'s really important) and they soon end up having a nasty fight. that\'s just the one main character.<br /><br />but lets get                                                 | 0.01          |\n| STAR RATING: \\*\\*\\*\\*\\*   | out of \\*\\*\\*\\*\\*\\* (I saw this in the theater when it was released and I think it really got worse after the theater screening.)<br /><br />RULES: You have to know what a "faux-" erotic film really is. And in the original, one of the female stars was not quite "comming"... It seemed like the director/author had tried to make his heroine the ultimate "sex object".<br /><br />WHAT IT IS: A sex                                                                          | 0.05          |\n| First off let             | me say that when I saw this movie I thought it was crap and it\'s definitely no worse then either it\'s trailer or any other. But on the other hand when I saw this movie and it started to get good with it\'s bad acting and the plot which was a bit weird and unoriginal and I almost cried but then it all started to get good. And some of the acting in some movies is good but none of the acting in this movie was that good and if not then at least                          | 0.11          |\n| I had high                | hopes for this movie since the reviews had me ready to get my money\'s worth. After being disappointed in the movie I wasn\'t expecting the review, I watched the video tape and now what do I think. You would think that since the title was so good that I would want to see it. What bothered me was the whole acting. It is not believable and the story is nothing new...like every movie that it is supposed to be very unrealistic. This movie was really hard to watch as the | 0.01          |\n| Isaac Florentine has      | created a very interesting character in this strange film. He\'s so different he\'s not interesting at all. I think this guy is the most fascinating and unique character in the entire movie.<br /><br />His life seems almost surreal. He works for a small restaurant, meets strange people. He plays the piano. And he\'s in love with a girl. So there you go. No dialogue, all a bizarre and surreal existence.<br /><br />One last note, the director of                         | 0.59          |\n| It actually pains         | me to say this, but as a film it was a horrible experience.<br /><br />From the time when they pulled away from the house and decided to pull over I got the impression that the lead role was the most serious actor in the film and was getting up to make some snappy lines to the lead guy with a couple of "I\'m still going to be in the car".<br /><br />I don\'t know that the writing was bad or poorly worded,                                                               | 0             |\n| Technically I\'am a        | great fan of Asian cinema, and I have seen some rather brilliant things in the last few years. There are however, movies that are both good and very bad.<br /><br />One example is \'Mulan\', in my opinion in my very opinion the world\'s greatest film. The film, by the way, is not as good as other Asian films such as \'A Man Called Thief\', \'The Red Snipers\', \'Man From Snowy River\', and \'Trouble                                                                             | 0.07          |\n| Honestly awful film,      | I don\'t think it\'s a good film at all if not. The acting was very poor too. What is the purpose of the story? The whole movie was confusing and silly. I got scared of it as soon as it came in the movie. Is it a horror movie like "Scary Movie", or is it a thriller like "Night of the Living Dead"? Probably a mixture but all it has is some horror elements and most of all a stupid movie.<br /><br />In                                                                     | 0             |\n| This flick is             | one of the most dull, slow and dull films I ever watched. The plot is silly & weak, the acting is terrible, bad special effects. The fact that you can even enjoy a movie as dull as this makes it even more dull, as it is so slow and boring. The only good thing about this movie is the soundtrack of some of the songs that are played in musical films such as "Stay With Me Until The Day Is Done", and "Shrek". The only thing that is                                       | 0             |\n| Blind Date (Columbia      | ) is a horror film that was released a good month after the much better \'Gremlins\'. It stars the talented cast of Julia Stiles, David Duchovny, Amanda Bynes, Marilu Henner, and John Carrott which is a shame because the screenplay is pretty good. However there are a few silly things in the script that make it hard to watch (I\'m saying that as an obsessive B-film fan). The plot centers around a man who is being                                                         | 0.19          |\n| I first watched           | this film about 30 years ago. I wasn\'t even a huge fan of the original cartoon series and was expecting no more than a silly animated film with a twist to the story (which was why it first aired). However, upon seeing it, I found myself laughing, I cried and so in spirit that I have no desire to re-watch this film. <br /><br />Firstly, why? because it stars a woman dressed up as her own alter ego in the story - it was                                                | 0.67          |\n| I saw the                 | film at the Venice Film Festival and was not impressed. It is obviously based on a work of Russian novelist Lev Solodovsky. If that is the inspiration you are looking for, the movie is not up to the task. Solodovsky did not manage to complete the novel in all its entirety, so his work does not lend itself this way. This is a disappointment. Solodovsky\'s novel includes no sex: neither erotic, nor asexual nor lesbian; one need not enjoy the novel\'s                   | 0.01          |\n| A group of                | friends and neighbors are terrorized in an abandoned amusement park by the psychotic Robert "the King Kid" Pyle. He uses a shotgun to shoot down the helpless victims. The police aren\'t called off till he has been apprehended and they call in an unlikely lead actress as "the Lady of Justice," who has escaped the clutches of an evil sheriff and set out to take revenge on Pyle and his band of murderous bullies.<br /><br />The "Lady of Justice" plot idea seems         | 0.05          |\n| Now, I LOVE               | that movie and I love that movie, but why bother watching this film which is actually one of the worst movies I have ever seen? This is one of the first films that I have ever laughed at seriously, and I\'m only 21. I love all the kids from my youth who have since gone on to so-called fame (Burt Reynolds, Ben Stiller, Jason Evers), and I love that they are now trying to market themselves (Burt Riedels, Ben Stiller                                                     | 0.33          |\n| This cheap, grainy-filmed | movie was made on film stock that I suppose was more plastic than real. It\'s not even nearly as scary as those cheap monster movies that make huge use of CGI - those that make use of special effects and a plot that can\'t have been written by an eleven-year-old. I have seen worse monster movies, and I\'ve seen some far worse action pictures, but the makers of this one can\'t even produce a good action movie. The makers of the "Ride with the Deadly                     | 0             |\n| I just finished           | watching the new remake of "Muppet Babies"--I can\'t believe that a movie, with so many potential stars, is trashed by such a film snobbishly dubbed. I saw the original movies before it was dubbed--it had the charm and power of the original. Granted, the movie was made in 1946 when they had their "Muppets" for the first time, but this movie was just so badly translated that the whole film loses its purpose without the benefit of                                      | 0             |\n| Ben, (Rupert Grint),      | a college student whose life is not quite what it seems, goes out with his friends to New York City to get laid for himself so he can pursue his dreams of becoming a super hero in the future. Upon arriving, he gets an idea to start his own superhero costume.<br /><br />In the beginning the film is not very good. There are a few cheap shots of people in the streets and a couple of clips where the hero is in action. I have not seen this movie for                     | 0.02          |\n| Every movie I             | have ever seen has disappointed me with this being what makes the movie,but with an excellent cast like Chris Smith(who also directed),and a decent script(this movie had a good story line and I believe the best and most funny part is when Chris finds himself in LA telling tales of his time in L.A.), this movie might not be the greatest movie I have seen but it is the best. If you can see the past of the movie,it is an action/comedy.                                 | 0.99          |\n\n</details>\n\nNote: if you try to plot a correlation plot between the target reward and the actual reward, it may look like it doesn\'t work well between the range of target reward (0.1, 0.9) . This is probably because the training dataset is heavily skewed towards 0.0 or 1.0 rewards.\n<details>\n  <summary>See correlation plot</summary>\n\n![correlation.png](eval_results%2Flarge_results%2Fcorrelation.png)\n</details>\n\n\n\n## How it works - details\n\nWe reserve the first token to encode the target reward. Or rather, the first unmasked token. \n\nThis means that when we pass input ids to the model for training, you need to make sure that you only pass model_max_length - 1 tokens.  \nThis library *should* handle the details of this happening.\n\n\nNOTE: Another way of doing this is to literally encode the reward as text input. A downside of this is that you\'ll probably be more open to prompt injection. [I demonstrate it here](https://github.com/thejaminator/prompt_reward_rl/blob/main/documentation/main_page.md#ability-to-match-a-single-reward)\nAnd you\'ll need to be more careful with how your rewards can get tokenized into multiple different tokens.\n\n\n\n## TODO list\n- [x] Validate that it works on a toy example\n- [ ] Reach out to others and ask if the hack makes sense\n- [x] Add support for huggingface pretrained models saving\n- [ ] Add collab notebook for toy example\n- [ ] Add examples for RLHF tasks - e.g. Openai\'s summarization where an [existing reward model is already available](https://huggingface.co/OpenAssistant)\n- [ ] Add support for some other pretrained models - not just gpt2\n- [ ] Write docs on how to add support for arbitrary pretrained models that are not added yet.\n- [ ] Settings for prompt vs completion token loss\n- [ ] Add support for online training\n',
    'author': 'James Chua',
    'author_email': 'chuajamessh@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/thejaminator/conditionme',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List

import torch
from datasets import Dataset
from transformers import (
    AutoTokenizer,
    GPT2LMHeadModel,
    AutoModelForCausalLM,
    TrainingArguments,
    Trainer,
    DataCollatorForLanguageModeling,
)

from conditionme.modified_gpt2_lm_head import ModifiedGPT2LMHeadModel
from conditionme.rollout.rollout_model import (
    complete_text_with_reward_batched,
    PromptCompletion,
)
from conditionme.cond_gpt2_tokenize import batch_tokenize_gpt2


def test_gpt_sanity():
    # create fake dataset for huggingface dataset
    dataset = {
        "text": [
            "This is a test",
            "This is another test",
            "This is a third test" * 1000,
        ],
        "target_reward": [0.1, 0.2, 0.3],
    }
    huggingface_dataset: Dataset = Dataset.from_dict(dataset)
    tokenizer = AutoTokenizer.from_pretrained(
        "sshleifer/tiny-gpt2", padding_side="left"
    )
    eos_token: str = tokenizer.eos_token
    tokenizer.pad_token = tokenizer.eos_token

    # tokenize the dataset
    dataset_tokenized = huggingface_dataset.map(
        lambda examples: batch_tokenize_gpt2(
            text=examples["text"],
            target_rewards=examples["target_reward"],
            tokenizer=tokenizer,
            add_eos_at_end=True,
        ),
        batched=True,
        batch_size=3,
    )
    dataset_tokenized.set_format(
        type="torch",
        columns=[
            "input_ids",
            "target_reward",
            "attention_mask",
        ],
    )
    device: torch.device = torch.device("cpu")
    # Train the model using the device
    gpt2_model: GPT2LMHeadModel = AutoModelForCausalLM.from_pretrained(
        "sshleifer/tiny-gpt2"
    ).to(device)
    model = ModifiedGPT2LMHeadModel.from_loaded_pretrained_model(
        loaded_model=gpt2_model
    )

    # Optionally save to drive
    training_args = TrainingArguments(
        output_dir="test_gpt_sanity",
        overwrite_output_dir=True,
        num_train_epochs=1,
        per_device_train_batch_size=16,
        save_steps=1,
        save_total_limit=2,
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=dataset_tokenized,
        data_collator=DataCollatorForLanguageModeling(tokenizer=tokenizer, mlm=False),
    )
    trainer.train()
    model.save_pretrained("test_gpt_sanity")
    new_model = ModifiedGPT2LMHeadModel.from_pretrained("test_gpt_sanity")


def test_complete_text_with_reward_batched():
    # create fake dataset for huggingface dataset
    dataset = {
        "text": [
            "This is a test",
            "This is another test",
            "This is thirdthirdthirdthird test test test",
        ],
        "target_reward": [0.1, 0.2, 0.3],
    }
    device: torch.device = torch.device("cpu")
    # Train the model using the device
    gpt2_model: GPT2LMHeadModel = AutoModelForCausalLM.from_pretrained(
        "sshleifer/tiny-gpt2"
    ).to(device)
    model = ModifiedGPT2LMHeadModel.from_loaded_pretrained_model(
        loaded_model=gpt2_model
    )
    huggingface_dataset: Dataset = Dataset.from_dict(dataset)
    tokenizer = AutoTokenizer.from_pretrained("sshleifer/tiny-gpt2")
    # Take 500 test set
    # convert into a list of space separated tokens
    test_text_tokenized: List[List[str]] = [
        text.split(" ") for text in huggingface_dataset["text"]
    ]
    # take the first 3 tokens from each list
    first_3_tokens_list: List[List[str]] = [text[:3] for text in test_text_tokenized]
    # join the first 3 tokens into a string
    first_3_tokens: List[str] = [" ".join(text) for text in first_3_tokens_list]
    # complete the text
    completions: List[PromptCompletion] = complete_text_with_reward_batched(
        prompt=first_3_tokens,
        model=model,
        tokenizer=tokenizer,
        target_reward=[1.0] * len(first_3_tokens),
        max_new_tokens=1,
    )
    assert len(completions) == len(test_text_tokenized)



import torch

from conditionme.reward_handler import DefaultRewardHandler, find_reward_token_position


def test_find_reward_token_position():
    input_ids = torch.LongTensor(
        [
            [5, 12312],
            [5, 2313],
            [5, 2],
        ],
    )
    reward_token_id = 5
    assert (
        find_reward_token_position(input_ids, reward_token_id).tolist()
        == torch.LongTensor([0, 0, 0]).tolist()
    )

    input_ids = torch.LongTensor([[5, 12312], [5, 2313], [2, 5]])
    reward_token_id = 5
    assert (
        find_reward_token_position(input_ids, reward_token_id).tolist()
        == torch.LongTensor([0, 0, 1]).tolist()
    )


def test_default():
    handler = DefaultRewardHandler(reward_token_id=5)
    # E.g. 3 examples in a batch
    target_reward = torch.tensor([1.0, 10.0, 100.0])
    # 3 x 2 x 2 hidden states
    hidden_states = torch.tensor(
        [
            # first sequence
            [
                [0.0, 1.0],  # first token
                [-1, -1],
            ],
            # second sequence
            [
                [0.0, 1.0],  # first token
                [-1, -1],
            ],
            # third sequence
            [
                [0.0, 1.0],  # first token
                [-1, -1],
            ],
        ]
    )
    # The reward token here are all the first tokens
    reward_token_id = 5
    input_ids = torch.LongTensor(
        [
            [5, 12312],
            [5, 2313],
            [5, 2],
        ],
    )

    modified_hidden_states = handler.handle_reward(
        target_reward,
        hidden_states,
        past_length=0,
        input_ids=input_ids,
    )
    assert modified_hidden_states[0, 0, -1] == 2.0, "should add 1.0 to the first token"
    assert (
        modified_hidden_states[1, 0, -1] == 11.0
    ), "should add 10.0 to the first token"
    assert (
        modified_hidden_states[2, 0, -1] == 101.0
    ), "should add 100.0 to the first token"


def test_default_different_reward_positions():
    handler = DefaultRewardHandler(reward_token_id=5)
    # E.g. 3 examples in a batch
    target_reward = torch.tensor([1.0, 1.0, 1.0])
    # 3 x 2 x 2 hidden states
    hidden_states = torch.tensor(
        [
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
            [
                [0.0, 0.0],
                [0.0, 0.0],
            ],
        ]
    )
    input_ids = torch.LongTensor(
        [
            [1, 5],  # First sequence has reward at second token,
            [5, 1],  # Second sequence has reward at first token
            [1, 5],  # Third sequence has reward at second token
        ],
    )
    modified_hidden_states = handler.handle_reward(
        target_reward=target_reward,
        hidden_states=hidden_states,
        input_ids=input_ids,
        past_length=0,
    )
    # first sequence
    assert (
        modified_hidden_states[0, 0, -1] == 0.0
    ), "should not add 1.0 to the first token"
    assert modified_hidden_states[0, 1, -1] == 1.0, "should add 1.0 to the second token"
    # second sequence
    assert modified_hidden_states[1, 0, -1] == 1.0, "should add 1.0 to the first token"
    # third sequence
    assert (
        modified_hidden_states[2, 0, -1] == 0.0
    ), "should not add 1.0 to the first token"
    assert modified_hidden_states[2, 1, -1] == 1.0, "should add 1.0 to the second token"



from copy import copy
from typing import Sequence, List

import torch
from transformers import (
    PreTrainedTokenizerBase,
    BatchEncoding,
    TensorType,
)
from transformers.utils import PaddingStrategy


def set_up_decoder_tokenizer(
    tokenizer: PreTrainedTokenizerBase,
) -> PreTrainedTokenizerBase:
    # shallow copy tokenizer to avoid unexpected side effects
    new_tokenizer: PreTrainedTokenizerBase = copy(tokenizer)
    # minus one to the sequence length to account for the reward token
    new_tokenizer.model_max_length = new_tokenizer.model_max_length - 1
    # need to manually set the pad token to the eos token
    new_tokenizer.pad_token = new_tokenizer.eos_token
    new_tokenizer.pad_token_id = new_tokenizer.eos_token_id
    # since this is a decoder, we need to left pad to work with HF generate
    # https://github.com/huggingface/transformers/issues/3021#issuecomment-1231526631
    new_tokenizer.padding_side = "left"
    # we also need to truncate from the left
    new_tokenizer.truncation_side = "left"
    return new_tokenizer


def manual_keep_front_truncation(
    input_ids: List[List[int]],
    max_length: int,
) -> List[List[int]]:
    # Because the transformers package truncates by keeping the last tokens, rather than the first tokens
    # we need to manually truncate the text
    # TODO: Check whether this is really true?
    return [row[:max_length] for row in input_ids]


def batch_tokenize_gpt2(
    text: Sequence[str],
    target_rewards: Sequence[float],
    tokenizer: PreTrainedTokenizerBase,
    add_eos_at_end: bool,
) -> BatchEncoding:
    # shallow copy tokenizer to avoid unexpected side effects
    new_tokenizer: PreTrainedTokenizerBase = set_up_decoder_tokenizer(tokenizer)
    assert len(text) == len(target_rewards)
    tokenized_ids = new_tokenizer(text)["input_ids"]
    # TODO: Do we still need to do manual truncation?

    # add reward_token to the start of all text, and add eos_token to the end of all text
    tokenized_ids_with_special_tokens: List[List[int]] = [
        row + ([new_tokenizer.eos_token_id] if add_eos_at_end else [])
        for row in tokenized_ids
    ]
    # Manually pad and truncate we want to add the token id ourselves
    tokenizer_result = new_tokenizer.pad(
        {
            "input_ids": manual_keep_front_truncation(
                tokenized_ids_with_special_tokens,
                max_length=new_tokenizer.model_max_length,
            )
        },
        padding=PaddingStrategy.LONGEST,
        return_attention_mask=True,
    )

    # BatchEncoding will have "input_ids", "attention_mask, "target_reward", "labels"
    # add the precomputed reward to the result
    tokenizer_result["target_reward"] = target_rewards
    # convert to tensors
    new_dict = BatchEncoding(tensor_type=TensorType.PYTORCH)
    for key in tokenizer_result:
        new_dict[key] = torch.tensor(tokenizer_result[key])
    return new_dict
